/*
 * sprawdzanie tautologii
 * (c) GPL Michal Moskal <malekith@pld.org.pl>
 */
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#define MAX_VARS ('z' - 'a' + 1)

/* symbols : */
enum {
t_error = 0,

t_open,		/*  (  */
t_close,	/*  )  */

t_first_2arg,
t_eq,		/* <-> ,=,==,<=> */
t_imp,		/*  -> ,>,=>     */
t_or,		/*  |  ,||       */
t_and,		/*  &  ,^,&&     */
t_last_2arg,

t_not,		/*  ~  ,-        */ /* only left is set, right == 0 */
t_false,	/*  0  ,F        */
t_true,		/*  1  ,T        */


t_var,		/* t_var + var_char - 'a' */
};

struct token
{
	int t, lev;
	struct token *left;
	struct token *right;
};

int used[MAX_VARS], vars[MAX_VARS];

int get()
{
	/* to be changed if needed */
	return getchar();
}

int exec(struct token *t)
{
	if (t == 0)
		abort();
		
	switch (t->t) {
	case t_not:
		return !exec(t->left);
	case t_and:
		return exec(t->left) && exec(t->right);
	case t_or:
		return exec(t->left) || exec(t->right);
	case t_eq:
		return exec(t->left) == exec(t->right);
	case t_imp:
		return !exec(t->left) || exec(t->right);
	case t_true:
		return 1;
	case t_false:
		return 0;
	default:
		return vars[t->t - t_var];
	}
}

static int one_char(int c)
{
	switch (c) {
	case '(':
		return t_open;
		
	case ')':
		return t_close;

	case '1':
	case 'T':
		return t_true;

	case '0':
	case 'F':
		return t_false;

	case '>':
		return t_imp;

	case '~':
		return t_not;

	case '^':
		return t_and;
	}
	
	return t_error;
}

/* this is lexical analyzer ... */
int get_symbol()
{
	static int last;
	int c, tmp;

	if (last) {
		c = last;
		last = 0;
	} else
		c = get();
		
	if (c == -1)
		return -1;

	while (isspace(c))
		c = get();
		
	if (c == -1)
		return -1;

	if ((tmp = one_char(c)) != t_error)
		return tmp;

	/* have to check the next one */
	last = get();
	
	switch (c) {
	case '-':
		if (last == '>') {
			last = 0;
			return t_imp;
		} else
			return t_not;
	case '=':
		if (last == '=') {
			last = 0;
			return t_eq;
		} else if (last == '>') {
			last = 0;
			return t_imp;
		} else
			return t_eq;

	case '|':
	case '&':
		if (last == c)
			last = 0;
		return c == '|' ? t_or : t_and;

	case '<':
		if (last != '=' && last != '-')
			return t_error;
		last = get();
		if (last != '>')
			return t_error;
		last = 0;
		return t_eq;
	}

	if (c >= 'a' && c <= 'z')
		return t_var + c - 'a';

	return t_error;
}

/* can be redefined to use some fixed space (array of tokens)
 * or sth */
struct token *new_token()
{
	struct token *r;

	r = (struct token*)malloc(sizeof(struct token));
	r->t = 0;
	r->left = r->right = 0;
	return r;
}

void croak(const char *msg)
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

/* this will make token from given symbol, will try to get next symbol
 * if x == '~' */

static int token_lev;

static struct token *make_token(int x)
{
	struct token *t;
	/* forward declaration to allow recursive calls between 
	 * make_token() and parse() */
	extern struct token *parse();

	if (x == t_open)
		return parse();
		
	t = new_token();
	t->t = x;
	t->lev = token_lev;
	
	switch (x) {
	case t_close:
		croak("unexpected `)'");
	case t_error:
		croak("lex error");
	case -1:
		croak("unexpected EOF");
	case t_not:
		t->left = make_token(get_symbol());
		break;
	}
	
	/* mark vars, so we know they were used */
	if (t->t >= t_var)
		used[t->t - t_var] = 1;
		
	return t;
}

static int is_lower(struct token *t, int t2)
{
	return (t->lev > token_lev) ||
	       (!(t->t == t2 && (t->t == t_eq || t->t == t_imp)) &&
	         (t->t >= t2));
}

/* will parse string till eos || ) || error */
struct token *parse()
{
	int x;
	struct token *head = 0, *p, *v;

	token_lev++;
	
	x = get_symbol();

	if (x == -1 || x == t_close)
		croak("unexpected EOF || )");

	head = make_token(x);

	for (;;) {
		x = get_symbol();

		if (x == -1 || x == t_close) {	/* only one thingie */
			token_lev--;
			return head;
		}

		if (!(x > t_first_2arg && x < t_last_2arg))
			croak("found !2arg");

		p = make_token(x);
		
		if (is_lower(head, p->t)) {
			p->left = head;
			head = p;
		} else {
			for (v = head; v->right; v = v->right)
				if (is_lower(v->right, p->t))
					break;
			p->left = v->right;
			v->right = p;
		}

		p->right = make_token(get_symbol());
	}
}

void destroy(struct token *t)
{
	if (t == 0)
		return;
	destroy(t->right);
	destroy(t->left);
	free(t);
}

void print(struct token *t)
{
	if (t->t > t_first_2arg && t->t < t_last_2arg) {
		printf("(");
		print(t->left);
		switch (t->t) {
		case t_or:
			printf(" | ");
			break;
		case t_and:
			printf(" & ");
			break;
		case t_eq:
			printf(" <=> ");
			break;
		case t_imp:
			printf(" => ");
			break;
		default:
			abort();
		}
		print(t->right);
		printf(")");
	} else
		switch (t->t) {
		case t_not:
			printf("~");
			print(t->left);
			break;
		case t_false:
		case t_true:
			printf("%d", t->t - t_false);
			break;
		default:
			printf("%c", t->t - t_var + 'a');
			break;
		}
}

/* increment vars[] according to used[],
 * ret -1 in case of end, 0 otherwise */
int inc()
{
	int i;

	for (i = 0; i < MAX_VARS; i++)
		if (used[i] == 1) {
			if (vars[i] == 0) {
				vars[i] = 1; 
				break;
			} else
				vars[i] = 0;
		}
				
	return (i == MAX_VARS) ? -1 : 0;
}

enum {
f_table 		= 0x0001,
f_print 		= 0x0002,
f_summary		= 0x0004,
};

int flags = 0xffff;
struct token *head;

void check()
{
	int r, t = 1, f = 1, i;

	if (flags & f_table) {
		for (i = 0; i < MAX_VARS; i++)
			if (used[i])
				printf("%c ", i + 'a');
		printf("W\n");
	}
	
	for (;;) {
		r = exec(head);
		if (r)
			f = 0;
		else
			t = 0;
		
		if (flags & f_table) {
			for (i = 0; i < MAX_VARS; i++)
				if (used[i])
					printf("%d ", vars[i]);
			printf("%d\n", r);
		}

		if (inc())
			break;
	}

	if (flags & f_summary) {
		if (t)
			printf("Tautologia.\n");
		else if (f)	
			printf("Sprzeczno.\n");
		else
			printf("Formua spenialna.\n");
	}
}

int main(int argc, char **argv)
{
	char *p;
	
	head = parse();

	if (argc > 1) 
		for (p = argv[1]; *p; p++)
			switch (*p) {
			case 't':
				flags &= ~f_table;
				break;
			case 'p':
				flags &= ~f_print;
				break;
			case 's':
				flags &= ~f_summary;
				break;
			}
			
	if (flags & f_print) {
		print(head);
		printf("\n");
	}

	check();
	
	destroy(head);
	return 0;
}
